#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <EzGraph.h>
#include "layout.h"
#include "bin_deg.h"
#include "main.h"
#include "game.h"
#include "snake.h"
#include "cell.h"
//#include "stage.h"
#include "draw.h"
#include "uty.h"

static game_state_t game_state;
static int dying_countdown;
static int clear_countdown;
static bin_deg_t turn_deg;
static int is_paused = 0;
static int hiscore, score;

game_state_t get_game_state(void){
    return game_state;
}
void set_game_state(game_state_t state){
    game_state = state;
    switch (state){
    case gs_normal:
        break;
    case gs_dying:
        dying_countdown = GAME_DYING_COUNTDOWN;
        break;
    case gs_clear:
        clear_countdown = GAME_CLEAR_COUNTDOWN;
        break;
    default:
        break;
    }
}


void game_init(void){
    set_hiscore(0);
    set_score(0);
}


void setup_game_classic(void){
    set_score(0);
    setup_cell();
    setup_obj_classic();
    setup_snake_classic(FIELD_WIDTH/2, FIELD_HEIGHT-FIELD_BORDER_MARGIN*2,
                        -BIN_DEG_RANGLE);
    //    setup_snake(160, 320, -BIN_DEG_RANGLE);
    //        setup_snake(160, 160, -BIN_DEG_RANGLE);
    //    setup_snake(160, 160, 0);
    EzSetKeyHandler(game_classic_keyh);
    EzSetKeyReleaseHandler(game_classic_key_releaseh);
    EzSetTimerHandler(game_classic_timerh, GAME_SPEED);

    set_game_state(gs_normal);
    //    setup_stage();
    setup_draw_screen();
    draw_screen();
    //    is_paused = 1;
    is_paused = 0;
}


void game_classic_keyh(int key){
    switch (key){
    case EZ_Left:
        turn_deg = -BIN_DEG_RANGLE;
        break;
    case EZ_Right:
        turn_deg = BIN_DEG_RANGLE;
        break;
    case EZ_Up:
        break;
    case EZ_Down:
        break;
    case 'p':
        if (game_state == gs_normal){
            if (is_paused){
                is_paused = 0;
            }
            else {
                is_paused = 1;
            }
        }
        break;
    case 'c':
        if (is_draw_cell_enabled()){
            disable_draw_cell();
        }
        else {
            enable_draw_cell();
        }
        draw_screen();
        break;
    default:
        break;
    }
}

void game_classic_key_releaseh(int key){
    switch (key){
    case EZ_Left:
    case EZ_Right:
        turn_deg = 0;
        break;
    case ' ':
        if (get_game_state() == gs_clear){
            clear_countdown = 0;
        }
        break;
    case 'g':
    case 'q':
        set_main_state(ms_title);        
        break;
    default:
        break;
    }
}

void game_classic_timerh(void){
    switch (game_state){
    case gs_normal:
        if (is_paused) return;
        reset_snake_bite();
        proceed_snake(turn_deg);
        set_snake_bite();
        if (judge_snake_bump()){
            set_game_state(gs_dying);
        }
        turn_deg = 0; // !!!とりあえず方向転換は単発押し。連続入力はキーリピートに頼る
        obj_iterator_act();
        break;

    case gs_dying:
        dying_countdown--;
        printf("dying ...%d\n", dying_countdown);
        if (dying_countdown < 0){
            set_main_state(ms_title);
        }
        break;
    case gs_clear:
        clear_countdown--;
        if (clear_countdown < 0){
            set_main_state(ms_title);
        }
    default:
        break;
    }
    draw_screen();
}

void setup_game_uroboros(void){
    set_score(0);
    setup_cell();
    setup_obj_uroboros();
    printf("setup end\n");
    setup_snake_uroboros(FIELD_WIDTH/2, FIELD_HEIGHT-FIELD_BORDER_MARGIN*2,
                         -BIN_DEG_RANGLE);
    //    setup_snake(160, 320, -BIN_DEG_RANGLE);
    //        setup_snake(160, 160, -BIN_DEG_RANGLE);
    //    setup_snake(160, 160, 0);
    EzSetKeyHandler(game_uroboros_keyh);
    EzSetKeyReleaseHandler(game_uroboros_key_releaseh);
    EzSetTimerHandler(game_uroboros_timerh, GAME_SPEED);

    set_game_state(gs_normal);
    //    setup_stage();
    setup_draw_screen();
    draw_screen();
    //    is_paused = 1;
    is_paused = 0;
}


void game_uroboros_keyh(int key){
    switch (key){
    case EZ_Left:
        turn_deg = -UROBOROS_TURN_DEG;
        break;
    case EZ_Right:
        turn_deg = UROBOROS_TURN_DEG;
        break;
    case EZ_Up:
        break;
    case EZ_Down:
        break;
    case 'p':
        if (game_state == gs_normal){
            if (is_paused){
                is_paused = 0;
            }
            else {
                is_paused = 1;
            }
        }
        break;
    case 'c':
        if (is_draw_cell_enabled()){
            disable_draw_cell();
        }
        else {
            enable_draw_cell();
        }
        draw_screen();
        break;
    default:
        break;
    }
}

void game_uroboros_key_releaseh(int key){
    switch (key){
    case EZ_Left:
    case EZ_Right:
        turn_deg = 0;
        break;
    case ' ':
        if (get_game_state() == gs_clear){
            clear_countdown = 0;
        }
        break;
    case 'g':
    case 'q':
        set_main_state(ms_title);        
        break;
    default:
        break;
    }
}

void game_uroboros_timerh(void){
    switch (game_state){
    case gs_normal:
        if (is_paused) return;
        reset_snake_bite();
        proceed_snake(turn_deg);
        set_snake_bite();
        if (judge_snake_bump()){
            snake_attack();
        }
        turn_deg = 0; // !!!とりあえず方向転換は単発押し。連続入力はキーリピートに頼る
        obj_iterator_act();
        break;

    case gs_dying:
        dying_countdown--;
        printf("dying ...%d\n", dying_countdown);
        if (dying_countdown < 0){
            set_main_state(ms_title);
        }
        break;
    case gs_clear:
        clear_countdown--;
        if (clear_countdown < 0){
            set_main_state(ms_title);
        }
    default:
        break;
    }
    draw_screen();
}

void quit_game(void){
}
void retry_game(void){
}

void set_score(int val){
    score = val;
}
int get_score(void){
    return score;
}
void add_score(int val){
    score += val;
    if (hiscore < score){
        hiscore = score;
    }
    update_draw_score();
}
void set_hiscore(int val){
    hiscore = val;
}
int get_hiscore(void){
    return hiscore;
}
