#include "config.h"
#include <stdio.h>
#include <string.h>
#include "obj.h"
#include "cell.h"
#include "rnd.h"
#include "game.h"
#include "snake.h"
//#include "draw.h"
#include "obj_draw.h"

void create_obj(obj_type_t type);
void obj_setup_egg(obj_t *obj);
void obj_act_egg(obj_t *obj);
void obj_setup_rat(obj_t *obj);
void obj_act_rat(obj_t *obj);
void obj_set_dest_rat(obj_t *obj);
void obj_setup_chickpea(obj_t *obj);
void obj_act_chickpea(obj_t *obj);
static void obj_draw_callback(obj_t *obj);
static void obj_act_callback(obj_t *obj);

static obj_info_t obj_infos[] = {
                          {oc_edible, ot_egg,
                           {"たまご", 32, 32},
                           {obj_setup_egg, obj_act_egg,
                            obj_draw_egg}},
                          {oc_enemy, ot_rat,
                           {"ねずみ", 32, 32},
                           {obj_setup_rat, obj_act_rat,
                            obj_draw_rat}},
                          {oc_edible, ot_chickpea,
                           {"ひよこまめ", 32, 32},
                           {obj_setup_chickpea, obj_act_chickpea,
                            obj_draw_chickpea}},


                          /*
                          {oc_edible, ot_chickpea, "ひよこまめ",
                           {NULL, NULL, NULL}},
                          {oc_enemy, ot_frog, "かえる",
                           {NULL, NULL, NULL}},
                          */
};

#define OBJ_INFO_NUM (sizeof(obj_infos)/sizeof(obj_info_t))

obj_funcs_t obj_funcs_table[ot_end];
obj_stats_t obj_stats_table[ot_end];

int obj_num;
obj_t objs[OBJ_NUM_MAX];

int chickpea_grow_counts[CHICKPEA_GROW_MAX] = {50, 50, 50, 50, 50};

typedef struct {
    double x, y;
    obj_t *obj;
} obj_event_t;

obj_event_t event_chickpea_grown;
#define IsEvent(event) ((event)->obj != NULL)
#define SetEvent(event, x, y, obj) \
    ((event)>x=x; (event)>y=y; (event)>obj=(obj))
#define ReSetEvent(event, x, y, obj) (obj=NULL)


void obj_callback_default(obj_t *obj){
}

void obj_init(void){
    int i;
    for (i = 0; i < OBJ_INFO_NUM; i++){
        obj_funcs_table[obj_infos[i].type] = obj_infos[i].funcs;
        obj_stats_table[obj_infos[i].type] = obj_infos[i].stats;
    }
    //    printf("%d->%s\n", ot_egg, obj_spec_table[ot_egg]->name);
    //    printf("%d->%s\n", ot_chickpea, obj_spec_table[ot_chickpea]->name);
}

void setup_obj(void){
    memset(objs, 0, sizeof(objs));
}


void setup_obj_classic(void){
    int i;
    setup_obj();
    for (i = 0; i < EGG_NUM; i++){
        create_obj(ot_egg);
    }
}

void setup_obj_uroboros(void){
    int i;
    setup_obj();
    for (i = 0; i < CHICKPEA_NUM; i++){
        create_obj(ot_chickpea);
    }
    for (i = 0; i < RAT_NUM; i++){
        create_obj(ot_rat);
    }
}

void create_obj(obj_type_t type){
    int i;
    for (i = 0; i < OBJ_NUM_MAX; i++){
        if (objs[i].valid) continue;
        printf("Creating %d:%s (%d) %p\n", type, obj_stats_table[type].name, i, &objs[i]);
        obj_funcs_table[type].setup(&objs[i]);
        objs[i].type = type;
        objs[i].valid = 1;
        obj_num++;
        break;
    }
    assert(i < OBJ_NUM_MAX);
}

void obj_iterator(obj_callback_t callback){
    int i;
    for (i = 0; i < obj_num; i++){
        callback(&objs[i]);
    }
}

void obj_iterator_draw(void){
    obj_iterator(obj_draw_callback);
}
void obj_iterator_act(void){
    obj_iterator(obj_act_callback);
}

static void obj_draw_callback(obj_t *obj){
    //    printf("draw callback %d:%s (%p)\n", obj->type, obj_stats_table[obj->type].name, obj);

    if (!obj->valid) return;
    assert(obj_funcs_table[obj->type].draw != NULL);
    obj_funcs_table[obj->type].draw(obj);
}

static void obj_act_callback(obj_t *obj){
    if (!obj->valid) return;
    assert(obj_funcs_table[obj->type].act != NULL);
    obj_funcs_table[obj->type].act(obj);
}


void obj_setup_egg(obj_t *obj){
    obj->x = rrnd(3, CELL_XNUM-3-1)*CELL_SIZE;
    obj->y = rrnd(3, CELL_YNUM-3-1)*CELL_SIZE;
}

void obj_act_egg(obj_t *obj){
    if (CountCellHeadFlag(obj->x, obj->y) > 2){
        printf("may be eaten!\n");
        if (is_head_overlapped(obj->x, obj->y, 4)){
            printf("eaten!\n");
            add_score(EGG_SCORE);
            obj_setup_egg(obj);
            grow_snake(EGG_FOOD_COUNT);
        }
    }
}

void obj_setup_chickpea(obj_t *obj){
    obj->x = rrnd(3, CELL_XNUM-3-1)*CELL_SIZE;
    obj->y = rrnd(3, CELL_YNUM-3-1)*CELL_SIZE;
    // ゲーム開始直後は成長具合をバラけさせる
    if (obj->ctr2 == 0){ 
        obj->ctr2 = rrnd(1, 3);
        obj->ctr1 = rrnd(0, chickpea_grow_counts[obj->ctr2]/2);
    }
    // ゲーム中の復活時はつねに初期状態からスタートする
    else {
        obj->ctr2 = obj->ctr1 = 0;
    }
}

void obj_act_chickpea(obj_t *obj){
    if (obj->ctr2 < CHICKPEA_GROW_MAX){
        obj->ctr1++;
        if (obj->ctr1 >= chickpea_grow_counts[obj->ctr2]){
            obj->ctr2++;
            obj->ctr1 = 0;
        }
    }
    //実がなっていなければ以下の捕食処理はパス
    if (obj->ctr2 < CHICKPEA_GROW_MAX) return;
    if (CountCellHeadFlag(obj->x, obj->y) > 2){
        printf("pea may be eaten!\n");
        if (is_head_overlapped(obj->x, obj->y, 16*16)){
            printf("pea eaten!\n");
            add_score(CHICKPEA_SCORE);
            obj_setup_chickpea(obj);
            grow_snake(CHICKPEA_FOOD_COUNT);
        }
    }
}

void obj_setup_rat(obj_t *obj){
    obj->x = rrnd(2, CELL_XNUM-3)*CELL_SIZE;
    obj->y = rrnd(2, CELL_YNUM-3)*CELL_SIZE;
    obj->ctr1 = 0;  // animation  counter
    obj->ctr2 = 0;  // respawn_counter;
    obj_set_dest_rat(obj);
}

void obj_act_rat(obj_t *obj){
    double new_x, new_y;
    // 死んだらカウンタが0になるまで活動しない
    if (obj->ctr2>0){
        obj->ctr2--;
        if (obj->ctr2 == 0) obj_setup_rat(obj);
        return;
    }
    obj->ctr1++;
    if (obj->ctr1==2) obj->ctr1 = 0;
    obj->deg = BinDegAtan(obj->dest_y-obj->y, obj->dest_x-obj->x);
    printf("rat (%.0f %.0f)->(%.0f %.0f) deg:%d\n",obj->x, obj->y, obj->dest_x, obj->dest_y, obj->deg);

    ResetCellRatFlag(obj->x, obj->y);
    new_x = obj->x + BinDegCos(obj->deg)*RAT_SPEED;
    new_y = obj->y + BinDegSin(obj->deg)*RAT_SPEED;

    // 蛇の胴体にぶつかったとき
    printf("rat new:(%.0f,%.0f)->(%d,%d)\n", new_x, new_y, X2Cx(new_x), Y2Cy(new_y));
    if (GetCellUnitCount(new_x, new_y)>0){
        // とりあえずその場で足踏み
        // 将来的には胴にそって移動する方向に転換する        
        new_x = obj->x;
        new_y = obj->y;
    }
    printf("rat moved\n");
    obj->x = new_x;
    obj->y = new_y;
    
    SetCellRatFlag(obj->x, obj->y);
    //目的地についたら再度目的地を設定
    if (is_overlapped(obj->x, obj->y, obj->dest_x, obj->dest_y, 8)){
        obj_set_dest_rat(obj);
    }
    // 蛇頭に重なっていたら蛇が死ぬ
    if (CountCellHeadFlag(obj->x, obj->y) > 2){
        printf("rat may be attack!\n");
        if (is_head_overlapped(obj->x, obj->y, 24*24)){
            printf("snake killed!!\n");
            set_game_state(gs_dying);
        }
    }
}


void obj_set_dest_rat(obj_t *obj){
    obj->dest_x = rrnd(2, CELL_XNUM-3)*CELL_SIZE;
    obj->dest_y = rrnd(2, CELL_YNUM-3)*CELL_SIZE;
}



int is_obj_overlapped(obj_t *obj1, obj_t *obj2, int dist2){
    return is_overlapped(obj1->x, obj1->y, obj2->x, obj2->y, dist2);
}

int is_overlapped(double x1, double y1, double x2, double y2, int dist2){
    printf("is_overlapped:(%.0f,%.0f)-(%.0f,%.0f) %.1f %d\n", x1, y1, x2, y2, (x2-x1)*(x2-x1)+(y2-y1)*(y2-y1), dist2);
           
    if ((x2-x1)*(x2-x1)+(y2-y1)*(y2-y1) <= dist2){
        return 1;
    }
    return 0;
}
